@@ -30,6 +30,7 @@ from plugins import GajimPlugin
 from plugins.helpers import log_calls
 from nbxmpp.simplexml import Node
 from nbxmpp import NS_CORRECT, NS_ADDRESS
+import nbxmpp
 
 from .xmpp import (
     NS_NOTIFY, NS_OMEMO, NS_EME, BundleInformationAnnouncement,
@@ -48,8 +49,7 @@ CRYPTOGRAPHY_MISSING = 'You are missing Python-Cryptography'
 AXOLOTL_MISSING = 'You are missing Python-Axolotl or use an outdated version'
 PROTOBUF_MISSING = 'OMEMO cant import Google Protobuf, you can find help in ' \
                    'the GitHub Wiki'
-GAJIM_VERSION = 'OMEMO only works with the latest Gajim version, get the ' \
-                'latest version from gajim.org'
+GAJIM_VERSION = 'Your Gajim version is to old, OMEMO needs 0.16.6 or higher'
 ERROR_MSG = ''
 
 NS_HINTS = 'urn:xmpp:hints'
@@ -86,7 +86,6 @@ if not ERROR_MSG:
         ERROR_MSG = 'Error: ' + str(e)
 
 GAJIM_VER = gajim.config.get('version')
-GROUPCHAT = False
 
 if os.name != 'nt':
     try:
@@ -98,18 +97,24 @@ if os.name != 'nt':
         ERROR_MSG = 'You are missing the Setuptools package.'
 
     if not SETUPTOOLS_MISSING:
-        if pkg.parse_version(GAJIM_VER) < pkg.parse_version('0.16.5'):
+        if pkg.parse_version(GAJIM_VER) < pkg.parse_version('0.16.6'):
             ERROR_MSG = GAJIM_VERSION
-        if pkg.parse_version(GAJIM_VER) > pkg.parse_version('0.16.5'):
-            GROUPCHAT = True
-else:
-    # if GAJIM_VER < 0.16.5, the Plugin fails on missing dependencys earlier
-    if not GAJIM_VER == '0.16.5':
-        GROUPCHAT = True
 
 # pylint: disable=no-init
 # pylint: disable=attribute-defined-outside-init
 
+ALLOWED_TAGS = [('request', nbxmpp.NS_RECEIPTS),
+                ('active', nbxmpp.NS_CHATSTATES),
+                ('gone', nbxmpp.NS_CHATSTATES),
+                ('inactive', nbxmpp.NS_CHATSTATES),
+                ('paused', nbxmpp.NS_CHATSTATES),
+                ('composing', nbxmpp.NS_CHATSTATES),
+                ('no-store', nbxmpp.NS_MSG_HINTS),
+                ('store', nbxmpp.NS_MSG_HINTS),
+                ('no-copy', nbxmpp.NS_MSG_HINTS),
+                ('no-permanent-store', nbxmpp.NS_MSG_HINTS),
+                ('replace', nbxmpp.NS_CORRECT),
+                ('thread', None)]
 
 class OmemoPlugin(GajimPlugin):
 
@@ -136,15 +141,15 @@ class OmemoPlugin(GajimPlugin):
             (ged.PRECORE, self.handle_outgoing_stanza),
             'message-outgoing':
             (ged.PRECORE, self.handle_outgoing_event)}
-        if GROUPCHAT:
-            self.events_handlers['gc-stanza-message-outgoing'] =\
-                (ged.PRECORE, self.handle_outgoing_gc_stanza)
-            self.events_handlers['gc-presence-received'] =\
-                (ged.PRECORE, self.gc_presence_received)
-            self.events_handlers['gc-config-changed-received'] =\
-                (ged.PRECORE, self.gc_config_changed_received)
-            self.events_handlers['muc-admin-received'] =\
-                (ged.PRECORE, self.room_memberlist_received)
+
+        self.events_handlers['gc-stanza-message-outgoing'] =\
+            (ged.PRECORE, self.handle_outgoing_gc_stanza)
+        self.events_handlers['gc-presence-received'] =\
+            (ged.PRECORE, self.gc_presence_received)
+        self.events_handlers['gc-config-changed-received'] =\
+            (ged.PRECORE, self.gc_config_changed_received)
+        self.events_handlers['muc-admin-received'] =\
+            (ged.PRECORE, self.room_memberlist_received)
 
         self.config_dialog = OMEMOConfigDialog(self)
         self.gui_extension_points = {'chat_control': (self.connect_ui,
@@ -573,11 +578,22 @@ class OmemoPlugin(GajimPlugin):
                 return
             if not state.encryption.is_active(to_jid):
                 return
-            # Delete previous Message out of Correction Message Stanza
-            if event.msg_iq.getTag('replace', namespace=NS_CORRECT):
-                event.msg_iq.delChild('encrypted', attrs={'xmlns': NS_OMEMO})
 
-            plaintext = event.msg_iq.getBody()
+            # Setting the ID here first for the old stanza object, which
+            # Gajim uses for message correction later.
+            # after cleanup_stanza() we have a new stanza object so we
+            # have to set the same ID so the ID we send matches with the
+            # ID from the correction stanza that Gajim hands us on a correction
+
+            # This is a nasty workaround, dont remove this or LMC in Groupchat
+            # will break for everything <= 0.16.8
+
+            new_id = gajim.connections[account].connection.getAnID()
+            event.msg_iq.setID(new_id)
+            self.cleanup_stanza(event)
+            event.msg_iq.setID(new_id)
+
+            plaintext = event.message
             msg_dict = state.create_gc_msg(
                 gajim.get_jid_from_account(account),
                 to_jid,
@@ -587,15 +603,13 @@ class OmemoPlugin(GajimPlugin):
 
             self.gc_message[msg_dict['payload']] = plaintext
             encrypted_node = OmemoMessage(msg_dict)
-            event.msg_iq.delChild('body')
             event.msg_iq.addChild(node=encrypted_node)
 
             # XEP-0380: Explicit Message Encryption
-            if not event.msg_iq.getTag('encryption', attrs={'xmlns': NS_EME}):
-                eme_node = Node('encryption', attrs={'xmlns': NS_EME,
-                                                     'name': 'OMEMO',
-                                                     'namespace': NS_OMEMO})
-                event.msg_iq.addChild(node=eme_node)
+            eme_node = Node('encryption', attrs={'xmlns': NS_EME,
+                                                 'name': 'OMEMO',
+                                                 'namespace': NS_OMEMO})
+            event.msg_iq.addChild(node=eme_node)
 
             # Add Message for devices that dont support OMEMO
             support_msg = 'You received a message encrypted with ' \
@@ -615,6 +629,8 @@ class OmemoPlugin(GajimPlugin):
             log.debug(e)
             return True
 
+        event.xhtml = None
+
     @log_calls('OmemoPlugin')
     def handle_outgoing_event(self, event):
         """ Handles a message outgoing event
@@ -671,12 +687,11 @@ class OmemoPlugin(GajimPlugin):
             if not state.encryption.is_active(to_jid):
                 return
 
-            # Delete previous Message out of Correction Message Stanza
-            if event.msg_iq.getTag('replace', namespace=NS_CORRECT):
-                event.msg_iq.delChild('encrypted', attrs={'xmlns': NS_OMEMO})
-
             plaintext = event.msg_iq.getBody().encode('utf8')
 
+            self.cleanup_stanza(event)
+            event._cb_parameters['msg_iq'] = event.msg_iq
+
             msg_dict = state.create_msg(
                 gajim.get_jid_from_account(account), to_jid, plaintext)
             if not msg_dict:
@@ -686,7 +701,6 @@ class OmemoPlugin(GajimPlugin):
 
             # Check if non-OMEMO resource is online
             contacts = gajim.contacts.get_contacts(account, to_jid)
-            non_omemo_resource_online = False
             for contact in contacts:
                 if contact.show == 'offline':
                     continue
@@ -697,18 +711,15 @@ class OmemoPlugin(GajimPlugin):
                     support_msg = 'You received a message encrypted with ' \
                                   'OMEMO but your client doesnt support OMEMO.'
                     event.msg_iq.setBody(support_msg)
-                    non_omemo_resource_online = True
-            if not non_omemo_resource_online:
-                event.msg_iq.delChild('body')
 
             event.msg_iq.addChild(node=encrypted_node)
 
             # XEP-0380: Explicit Message Encryption
-            if not event.msg_iq.getTag('encryption', attrs={'xmlns': NS_EME}):
-                eme_node = Node('encryption', attrs={'xmlns': NS_EME,
-                                                     'name': 'OMEMO',
-                                                     'namespace': NS_OMEMO})
-                event.msg_iq.addChild(node=eme_node)
+
+            eme_node = Node('encryption', attrs={'xmlns': NS_EME,
+                                                 'name': 'OMEMO',
+                                                 'namespace': NS_OMEMO})
+            event.msg_iq.addChild(node=eme_node)
 
             # Store Hint for MAM
             store = Node('store', attrs={'xmlns': NS_HINTS})
@@ -718,6 +729,21 @@ class OmemoPlugin(GajimPlugin):
             log.debug(e)
             return True
 
+        event.xhtml = None
+
+    @staticmethod
+    def cleanup_stanza(obj):
+        ''' We make sure only allowed tags are in the stanza '''
+        stanza = nbxmpp.Message(
+            to=obj.msg_iq.getTo(),
+            typ=obj.msg_iq.getType())
+        stanza.setThread(obj.msg_iq.getThread())
+        for tag, ns in ALLOWED_TAGS:
+            node = obj.msg_iq.getTag(tag, namespace=ns)
+            if node:
+                stanza.addChild(node=node)
+        obj.msg_iq = stanza
+
     @log_calls('OmemoPlugin')
     def handle_device_list_update(self, event):
         """ Check if the passed event is a device list update and store the new
@@ -1100,7 +1126,8 @@ class OmemoPlugin(GajimPlugin):
                 log.error(account + ' => Devicelistquery was NOT successful')
                 self.publish_own_devices_list(account, new=True)
                 return False
-            contact_jid = stanza.getAttr('from')
+
+            contact_jid = str(stanza.getAttr('from'))
             if contact_jid == my_jid:
                 state.set_own_devices(devices_list)
                 state.store.sessionStore.setActiveState(devices_list, my_jid)
